<?php

namespace Drupal\gridstack;

use Drupal\Component\Utility\Html;
use Drupal\gridstack\Entity\GridStack;

/**
 * Provides GridStack attribute utilities.
 */
class GridStackAttribute {

  /**
   * Returns the wrapper attributes.
   */
  public static function prepare($optionset, array $settings, $ungridstack = FALSE) {
    $attributes = empty($settings['attributes']) ? [] : self::parse($settings['attributes']);

    if ($ungridstack) {
      return $attributes;
    }

    // Adds wrapper classes for static grid Bootstrap/ Foundation, or js-driven.
    if (!empty($settings['wrapper_classes']) && is_string($settings['wrapper_classes'])) {
      self::parseClasses($attributes, $settings['wrapper_classes']);
    }
    if ($settings['use_js']) {
      // Adds attributes for js-driven layouts.
      // Gets options.breakpoints.sm.[width, column, image_style, grids], etc.
      $exclude_image_style = empty($settings['_admin']);
      $columns = $optionset->getJson('breakpoints');
      if ($responsives = array_filter($optionset->getBreakpoints())) {
        foreach (GridStack::getConstantBreakpoints() as $breakpoint) {
          $responsive_grids = $optionset->getJsonSummaryBreakpoints($breakpoint, $exclude_image_style);
          $has_width = isset($responsives[$breakpoint]['width']) && $responsives[$breakpoint]['width'] > -1;
          if ($has_width && $responsive_grids) {
            $attributes['data-' . $breakpoint . '-width'] = $responsives[$breakpoint]['width'];
            $attributes['data-' . $breakpoint . '-grids'] = $responsive_grids;
          }
        }
      }

      // Add the required configuration as JSON object.
      $attributes['data-breakpoints'] = strpos($columns, '{"":12}') !== FALSE ? '' : $columns;
      $attributes['data-config'] = $optionset->getJson('settings');

      // Breakpoint related data-attributes helpers.
      if (!empty($settings['minWidth'])) {
        $attributes['data-min-width'] = (int) $settings['minWidth'];
      }
    }

    return $attributes;
  }

  /**
   * Provides dynamic GridStack JS grid attributes.
   */
  public static function jsBox($optionset, array &$settings, $current = 'grids') {
    $attributes  = self::regionBox($settings, $current);
    $nameshort   = $settings['nameshort'];
    $breakpoints = isset($settings['breakpoints']) ? $settings['breakpoints'] : [];
    $keys        = $breakpoints ? array_keys($breakpoints) : [];
    $end_key     = $keys ? end($keys) : 'xl';
    $breakpoint  = isset($settings['breakpoint']) ? $settings['breakpoint'] : $end_key;
    $id          = isset($settings['delta']) ? $settings['delta'] : 0;
    $nid         = isset($settings['nested_delta']) ? $settings['nested_delta'] : NULL;
    $end_grids   = $optionset->getEndBreakpointGrids($current);
    $grids       = isset($breakpoints[$breakpoint]) && isset($breakpoints[$breakpoint][$current]) ? $breakpoints[$breakpoint][$current] : $end_grids;

    // Nested grids.
    if (isset($settings['nested_delta']) && $current == 'nested') {
      foreach (['x', 'y', 'width', 'height'] as $key) {
        if (!isset($grids[$id][$nid])) {
          continue;
        }

        $attributes['data-' . $nameshort . '-' . $key] = isset($grids[$id][$nid][$key]) ? (int) $grids[$id][$nid][$key] : 0;
      }
    }
    else {
      // The root element grids.
      foreach (['x', 'y', 'width', 'height'] as $key) {
        $attributes['data-' . $nameshort . '-' . $key] = isset($grids[$id][$key]) ? (int) $grids[$id][$key] : 0;
      }
    }

    return $attributes;
  }

  /**
   * Provides static Bootstrap/ Foundation CSS grid attributes.
   */
  public static function cssBox($optionset, array &$settings, $current = 'grids', $optimize = FALSE) {
    $framework  = $settings['framework'];
    $attributes = self::regionBox($settings, $current);
    $points     = GridStackDefault::breakpoints();

    // Bootstrap 4 uses flexbox with `col` class, and has `xl` breakpoint.
    if ($framework == 'bootstrap') {
      $attributes['class'][] = 'col';
    }
    elseif ($framework == 'foundation') {
      unset($points['xs'], $points['xl']);
    }

    $unique = $optionset->optimizeGridWidths($settings, $current, $optimize);
    foreach ($points as $point => $label) {
      if (!isset($unique[$point])) {
        continue;
      }

      $prefix = $suffix = '';
      if (strpos($framework, 'bootstrap') !== FALSE) {
        // Specific to XS: Bootstrap 3: col-xs-*, Bootstrap 4: col-*.
        $prefix = 'col-' . $point . '-';
        if ($framework == 'bootstrap' && $point == 'xs') {
          $prefix = 'col-';
        }
      }
      elseif ($framework == 'foundation') {
        $prefix = $label . '-';
        $suffix = ' columns';
      }

      $attributes['class'][] = $prefix . $unique[$point] . $suffix;
    }

    return $attributes;
  }

  /**
   * Parses the given string classes.
   */
  private static function parseClasses(array &$attributes, $string = '') {
    $classes = array_map('\Drupal\Component\Utility\Html::cleanCssIdentifier', explode(' ', $string));
    $attributes['class'] = empty($attributes['class']) ? array_unique($classes) : array_unique(array_merge($attributes['class'], $classes));
  }

  /**
   * Provides both CSS grid and js-driven attributes configurable via UI.
   */
  private static function regionBox(array &$settings, $current = 'grids') {
    $id         = isset($settings['delta']) ? $settings['delta'] : 0;
    $nid        = isset($settings['nested_delta']) ? $settings['nested_delta'] : NULL;
    $regions    = isset($settings['regions']) ? $settings['regions'] : [];
    $rid        = $current == 'nested' ? 'gridstack_' . $id . '_' . $nid : 'gridstack_' . $id;
    $region     = empty($regions[$rid]) ? [] : $regions[$rid];
    $attributes = [];

    if ($region) {
      if (isset($region['attributes']) && !empty($region['attributes'])) {
        $attributes = self::parse($region['attributes']);
        unset($settings['regions'][$rid]['attributes']);
      }

      if (!empty($region['wrapper_classes'])) {
        self::parseClasses($attributes, $region['wrapper_classes']);
        unset($settings['regions'][$rid]['wrapper_classes']);
      }
    }

    return $attributes;
  }

  /**
   * Parses the given string attribute.
   */
  private static function parse($string = '') {
    $attributes = [];
    // Given role|navigation,data-something|some value.
    $layout_attributes = explode(',', $string);
    foreach ($layout_attributes as $attribute) {
      $replaced_attribute = $attribute;

      // @nottodo: Token support.
      // No need to whitelist as this already requires admin priviledges.
      // With admin privileges, the site is already taken over before playing
      // around with attributes. However provides few basic sanitizations to
      // satisfy curious playful editors.
      if (strpos($attribute, '|') !== FALSE) {
        list($key, $value) = array_pad(array_map('trim', explode('|', $replaced_attribute, 2)), 2, NULL);
        $key = mb_substr($key, 0, 2) === 'on' ? 'data-' . $key : $key;
        $attributes[$key] = Html::cleanCssIdentifier(strip_tags($value));
      }
    }

    return $attributes;
  }

}
